function [] = Ex_28p2_004_polynomial()
% ---------------------------------------------------------------------------
% Exercise 28.2 in MacKay's book using fun_solve_poly.m 
% by Matthew Yoko and Matthew Juniper 2022
% ---------------------------------------------------------------------------
% PRIOR READING
%  Information Theory, Inference, and Learning Algorithms, by David J.C.
%  MacKay, Cambridge University Press (v7.2 2005),
%  https://www.inference.org.uk/itprnn/book.pdf 
% 
% Essential reading:
%  Chapters  1 -  3; 
%  Chapters 27 - 28.
%
% Useful reading:
%  Chapters 21, 22, 24.
%  Exercise 28.1
% ---------------------------------------------------------------------------

% ---------------------------------------------------------------------------
% QUESTION
% ---------------------------------------------------------------------------
% We are given data, D, and asked to evaluate the evidence (a.k.a. marginal
% likelihood) for two models, H1(w) and H2(w), given the data, D, where
% w are the models' parameters. In other words:
% 
%   Calculate the evidences, P(D|H1) and P(D|H2), where:
% 
%   P(D|Hi) = int{ P(D|w,Hi) * P(w|Hi) } dw  (Eq. 28.7) 
% 
%   D is [x] = [ -8 , -2 ,  6 ]
%        [t]   [  8 , 10 , 11 ]
%
%   The measurement, t, is y + Normal(0,1)
% 
%   Model H1 is y = w0
%   Model H2 is y = w0 + w1*x
% 
%   Prior w0 is Normal(0,1) for both models
%   Prior w1 is Normal(0,1)

% ---------------------------------------------------------------------------
% Set the user-defined variables
% ---------------------------------------------------------------------------
%
% Enter the data from Exercise 28.2 in MacKay's book
x_e = [ -8 , -2 ,  6 ];
y_e = [  8 , 10 , 11 ];
% Set the measurement variance
C_ee = 1;
% Enter the prior expected values and variances of the parameters w0 and w1
w0_E = 0.0; w0_V = 1.0;
w1_E = 0.0; w1_V = 1.0;

% ---------------------------------------------------------------------------
% SOLUTION
% ---------------------------------------------------------------------------
% We use Laplace's method, as described in Ex_28p2_002_Laplaces_method.m .
%
% Model H1
[w0_MAP_H1,inv_C_MAP_H1,log_P_D_H1,~,~] = fun_MAP(0,w0_E,1/w0_V,1/C_ee,x_e,y_e);
% Model H2
[wn_MAP_H2,inv_C_MAP_H2,log_P_D_H2,~,~] = fun_MAP(1,[w0_E,w1_E],[1/w0_V,0;0,1/w1_V],1/C_ee,x_e,y_e);

% ---------------------------------------------------------------------------
% RESULTS THAT CAN BE COMPARED DIRECTLY WITH PREVIOUS CODES
% ---------------------------------------------------------------------------
fprintf('------------------------------------------------ \n')
fprintf(' log(evidence) calculated with fun_MAP_poly.m    \n')
fprintf('------------------------------------------------ \n')
fprintf('log{ P(D|H1) } = %6.4f \n',log_P_D_H1)
fprintf('log{ P(D|H2) } = %6.4f \n',log_P_D_H2)
fprintf('------------------------------------------------ \n')
fprintf('P(D|H1) / P(D|H2) = %6.4f \n',exp(log_P_D_H1 - log_P_D_H2))
fprintf('------------------------------------------------ \n')
fprintf(' MAP point calculated with fun_MAP_poly.m        \n')
fprintf('------------------------------------------------ \n')
fprintf('max{ P(D|H1) } occurs at w = ( %6.4f )   \n',w0_MAP_H1)
fprintf('max{ P(D|H2) } occurs at w = ( %6.4f , %6.4f ) \n',wn_MAP_H2(2),wn_MAP_H2(1))
fprintf('------------------------------------------------ \n')

% ---------------------------------------------------------------------------
% DISCUSSION
% ---------------------------------------------------------------------------
% These results are the same as those calculated previously but use the
% more general function fun_MAP.m  

% ---------------------------------------------------------------------------
% CALCULATING POSTERIOR UNCERTAINTIES IN THE MODEL PREDICTIONS
% ---------------------------------------------------------------------------
% Having calculated the inverse of the posterior covariance of the parameters, inv_C_MAP_Hi, we have:
%   p(a|D,H) ~ exp{ - 0.5 * (a - a0)' * inv_C_MAP_Hi * (a - a0) } 
% where a0 is the most likely set of parameters and C_MAP_Hi is the posterior
% covariance matrix around this solution.
%
% We want to find the corresponding confidence intervals in the model
% predictions. (see § 4.3 of Hans Yu's thesis).
%
% We want to find a for which |y(a) - y(a0)|^2 is maximized, subject to:
% (a - a0)' * inv(C_aa_post) * (a - a0) = r^2, where r denotes the
% confidence level in units of standard deviations. 
%
% A first order approximation of the above gives:
%  |y(a) - y(a0)|^2 ~ (a - a0)' * [dyda * dyda'] * (a - a0)
%                                 <-outer prod->
%
% The method of Lagrange multipliers gives the necessary condition:
%  [dyda * dyda'] * (a - a0) = lambda * inv_C_MAP_Hi * (a - a0)
% 
% This is a generalized eigenvalue problem for lambda:
%  [lambda , (a - a0)] = eig( [dyda * dyda'] , inv_C_MAP_Hi )
%
% The standard deviation of y is given by the maximum eigenvalue, lambda

% Set the values of x at which the model predictions and their uncertainties will be plotted
x_plt = -8:1:+8; N = numel(x_plt);
% Initialize
E_y_post_H1 = NaN(1,N); V_y_post_H1 = NaN(1,N);
E_y_post_H2 = NaN(1,N); V_y_post_H2 = NaN(1,N);
% Loop through the points to calculate the uncertainties in the model predictions 
for n = 1:numel(x_plt)
    % Calculate y and dyda at these param values
    [temp_H1,dyda_H1,~] = fun_solve_poly(w0_MAP_H1,x_plt(n)); 
    [temp_H2,dyda_H2,~] = fun_solve_poly(wn_MAP_H2,x_plt(n)); 
    % Extract the expected values from the model
    E_y_post_H1(n) = temp_H1;
    E_y_post_H2(n) = temp_H2;
    % Extract the variance from the model
    V_y_post_H1(n) = max(eig(dyda_H1 * dyda_H1.',inv_C_MAP_H1));
    V_y_post_H2(n) = max(eig(dyda_H2 * dyda_H2.',inv_C_MAP_H2));
end
% Plot these
figure(1); clf; hold on; RGB = gen_RGB;
% Plot model 1 predictions with errorbars
[han_H1_E,~] = gen_errorpatch(x_plt,E_y_post_H1,V_y_post_H1,RGB.blue);
% Plot model 2 predictions with errorbars
[han_H2_E,~] = gen_errorpatch(x_plt,E_y_post_H2,V_y_post_H2,RGB.orange);
% Plot the original data
han_D        = gen_errorbar(x_e,y_e,2*sqrt(C_ee)*ones(1,numel(x_e)),'ko','MarkerFaceColor','k');
% Annotate the plot
legend([han_H1_E,han_H2_E,han_D],{'Model 1 predictions ± 2 s.d.','Model 2 predictions ± 2 s.d','Data'},'Location','SouthEast'); xlabel('x'); ylabel('y'); title('Data and model predictions at their MAP points, showing ± 2 s.d.')

end


