function [] = Sec_001()
%--------------------------------------------------------------------------
% Fit a polynomial to data generated from an nth order polynomial with fixed measurement noise 
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% Set the model (a.k.a. the order of the polynomial)
%--------------------------------------------------------------------------
N_mod = 5;

%--------------------------------------------------------------------------
% Set the plot options
%--------------------------------------------------------------------------
plot_options.plot_gt   = true; % plot ground truth polynomial
plot_options.plot_data = true; % plot data 
plot_options.plot_pred = true; % plot model predictions

% Open a figure and initialize plot_options if they are required
if plot_options.plot_gt || plot_options.plot_data || plot_options.plot_pred
    figure(1); clf; hold on
    plot_options.plot_handles = [];
    plot_options.plot_labels  = {};
end
% Initialize x_plt if it is required
if plot_options.plot_gt || plot_options.plot_pred
    x_plt = -0.5:0.01:0.5; n_plt = numel(x_plt);
end

%--------------------------------------------------------------------------
% Generate ground truth parameters, a_gt, from a polynomial of a user-defined order
%--------------------------------------------------------------------------
% Set the order of the ground truth polynomial
N_gt    = 5;
% Reset the random seed (optional)
rng(1);
% Generate some random data 
x_ref = -0.5:0.05:0.5; n_ref = numel(x_ref); y_ref = randn(n_ref,1);
% Find the coefficients of an order N_gt polynomial that fits this data
a_gt = polyfit(x_ref,y_ref,N_gt);
% Plot ground truth
if plot_options.plot_gt
    new_handle = plot(x_plt,polyval(a_gt,x_plt),'r-');
    new_label  = 'ground truth';
    plot_options = fun_update_legend(plot_options,new_handle,new_label);
end

%--------------------------------------------------------------------------
% Generate N_exp synthetic datapoints from the ground truth polynomial
%--------------------------------------------------------------------------
%
% Set the number of experimental datapoints
N_e  = 10;
% Set the x-locations at which the datapoints are selected
x_e  = linspace(-0.5,0.5,N_e);
% Set the variance of the measurement noise
C_ee = 0.0005;
% Sample the polynomial at the measurement points, corrupted by Gaussian noise 
y_e  = polyval(a_gt,x_e) + sqrt(C_ee)*randn(1,N_e);
% Plot data
if plot_options.plot_data
    new_handle = errorbar(x_e,y_e,2*sqrt(C_ee)*ones(N_e,1),'k.');
    new_label  = 'experimental data';
    plot_options = fun_update_legend(plot_options,new_handle,new_label);
end

%--------------------------------------------------------------------------
% Set the prior and the prior covariances
%--------------------------------------------------------------------------
% 
% The prior and the prior covariances have a strong influence on the
% marginal likelihood. For example, see MacKay (1992) on how to compare
% models with different priors. 
%
% Here, we will set the priors expected values and variances to those found
% from Monte Carlo sampling of the above ground truth parameters, performed
% in prior_MC.m. This is a brute force solution, but is quick. It gives:
%
%  mean = zero
%  var  ~ exp(2*(N_mod:-1:0)), where N_mod is the model polynomial's order
% 
% Other covariances can be set with the switch below. 
%
% MacKay (1992) 'Bayesian Interpolation'; Neural Computation 4, 415--447

% Set the type of prior
prior = 'zero';
% Set the type of covariance
prior_covariance = 'std_propto_exp_n';

% Set the prior
switch prior
    case 'zero'
        a_prior = zeros(1,N_mod+1);
    otherwise
        disp('prior not set'); return
end

% Set the covariance
switch prior_covariance
    case 'std_propto_exp_n'
        % Approx. match to MC sample of priors
        C_aa_prior = diag(exp(2*(N_mod:-1:0)));
    case 'propto_n'
        % covariance scales with 2^(n), where n is the parameter multiplied by x^n. 
        C_aa_prior = 2 * diag(2.^(N_mod:-1:0));
    case 'loose_propto_n'
        C_aa_prior = 2^4 * diag(2.^(N_mod:-1:0));
    case 'very_loose_propto_n'
        C_aa_prior = 2^8 * diag(2.^(N_mod:-1:0));
    case 'unity'
        % covariance is the same for all coefficients
        C_aa_prior = eye(N_mod+1);
    otherwise
        disp('prior covariance not set'); return
end

%--------------------------------------------------------------------------
% Calculate the MAP points, the Marginal Likelihood, the BFL and the OF
%--------------------------------------------------------------------------
%
[a_MAP,inv_C_aa_MAP,log_ML,log_BFL,log_OF] = fun_MAP(N_mod,a_prior,inv(C_aa_prior),1/C_ee,x_e,y_e);

% ---------------------------------------------------------------------------
% Print results to screen
% ---------------------------------------------------------------------------
fprintf('------------------------------------------------ \n')
fprintf(' Model is polynomial of order %u                 \n',N_mod)
fprintf('------------------------------------------------ \n')
fprintf('log_ML  = log{ P(D|H) }       = %6.4f \n',log_ML )
fprintf('log_BFL = log{ P(D|A_MAP,H) } = %6.4f \n',log_BFL)
fprintf('log_OF  =                     = %6.4f \n',log_OF )
fprintf('------------------------------------------------ \n')

%--------------------------------------------------------------------------
% Calculate the poterior uncertainties in the model predictions
%--------------------------------------------------------------------------
if plot_options.plot_pred
    % Initialize E and V
    E_y_post = NaN(1,n_plt); V_y_post = NaN(1,n_plt);
    % Loop through the points to calculate the uncertainties in the model predictions
    for n = 1:numel(x_plt)
        % Calculate y and dyda at these param values
        [temp,dyda,~] = fun_solve_poly(a_MAP,x_plt(n));
        % Extract the expected values from the model
        E_y_post(n) = temp;
        % Extract the variance from the model
        V_y_post(n) = max(eig(dyda * dyda.',inv_C_aa_MAP));
    end
    % Plot these
    RGB = gen_RGB;
    % Plot model 1 predictions with errorbars
    han_E = gen_errorpatch(x_plt,E_y_post,V_y_post,RGB.blue);
    plot_options = fun_update_legend(plot_options,han_E,'Model prediction ± 2 s.d.');
    title('Calculated with polynomial assimilation code')
end

%--------------------------------------------------------------------------
% Plot to screen or file
%--------------------------------------------------------------------------
if plot_options.plot_gt || plot_options.plot_data || plot_options.plot_pred
    legend(plot_options.plot_handles,plot_options.plot_labels)
    xlabel('x'); ylabel('y'); drawnow
end

end

%--------------------------------------------------------------------------
% Utility functions
%--------------------------------------------------------------------------

function [plot_options] = fun_update_legend(plot_options,new_handle,new_label)

% Update the legend markers, handles, and labels
plot_options.plot_handles(end+1) = new_handle;
plot_options.plot_labels{end+1}  = new_label;

end



