function [] = Sec_002()
%--------------------------------------------------------------------------
% Fit several polynomials to data generated from an nth order polynomial with fixed measurement noise 
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
% Generate ground truth parameters, a_gt, from a polynomial of a user-defined order
%--------------------------------------------------------------------------
% Set the order of the ground truth polynomial
N_gt    = 5;
% Reset the random seed (optional)
rng(1);
% Generate some random data 
x_ref = -0.5:0.05:0.5; n_ref = numel(x_ref); y_ref = randn(n_ref,1);
% Find the coefficients of an order N_gt polynomial that fits this data
a_gt = polyfit(x_ref,y_ref,N_gt);

%--------------------------------------------------------------------------
% Generate N_exp synthetic datapoints from the ground truth polynomial
%--------------------------------------------------------------------------
%
% Set the number of experimental datapoints
N_e  = 10;
% Set the x-locations at which the datapoints are selected
x_e  = linspace(-0.5,0.5,N_e);
% Set the variance of the measurement noise
C_ee = 0.0005;
% Sample the polynomial at the measurement points, corrupted by Gaussian noise 
y_e  = polyval(a_gt,x_e) + sqrt(C_ee)*randn(1,N_e);

%--------------------------------------------------------------------------
% Set the type of prior and the type of prior covariances
%--------------------------------------------------------------------------
% (See comments in Sec_001.m)
%
% Set the type of prior
prior = 'zero';
% Set the type of covariance
prior_covariance = 'std_propto_exp_n';

%--------------------------------------------------------------------------
% Loop over polynomials from 1 to 9
%--------------------------------------------------------------------------
% Pre-allocate log_ML, log_BFL, log_OF
log_ML  = NaN(1,9);
log_BFL = NaN(1,9);
log_OF  = NaN(1,9);

for N_mod = 1:9

    % Set the prior
    switch prior
        case 'zero'
            a_prior = zeros(1,N_mod+1);
        otherwise
            disp('prior not set'); return
    end

    % Set the covariance
    switch prior_covariance
        case 'std_propto_exp_n'
            % Approx. match to MC sample of priors 
            C_aa_prior = diag(exp(2*(N_mod:-1:0)));
        case 'propto_n'
            % covariance scales with 2^(n), where n is the parameter multiplied by x^n.
            C_aa_prior = 2 * diag(2.^(N_mod:-1:0));
        case 'loose_propto_n'
            C_aa_prior = 2^4 * diag(2.^(N_mod:-1:0));
        case 'very_loose_propto_n'
            C_aa_prior = 2^8 * diag(2.^(N_mod:-1:0));
        case 'unity'
            % covariance is the same for all coefficients
            C_aa_prior = eye(N_mod+1);
        otherwise
            disp('prior covariance not set'); return
    end

    %--------------------------------------------------------------------------
    % Calculate the MAP points, the Marginal Likelihood, the BFL and the OF
    %--------------------------------------------------------------------------
    %
    [~,~,log_ML_,log_BFL_,log_OF_] = fun_MAP(N_mod,a_prior,inv(C_aa_prior),1/C_ee,x_e,y_e);
    log_ML(N_mod)  = log_ML_;
    log_BFL(N_mod) = log_BFL_;
    log_OF(N_mod)  = log_OF_;    

    % ---------------------------------------------------------------------------
    % Print results to screen
    % ---------------------------------------------------------------------------
    fprintf('------------------------------------------------ \n')
    fprintf(' Model is polynomial of order %u                 \n',N_mod)
    fprintf('------------------------------------------------ \n')
    fprintf('log_ML  = log{ P(D|H) }       = %6.4f \n',log_ML_ )
    fprintf('log_BFL = log{ P(D|A_MAP,H) } = %6.4f \n',log_BFL_)
    fprintf('log_OF  =                     = %6.4f \n',log_OF_ )
    fprintf('------------------------------------------------ \n')

end

%--------------------------------------------------------------------------
% Plot to screen or file
%--------------------------------------------------------------------------
figure(1); clf
subplot(3,1,1); plot(1:9,log_BFL,'o-'); grid on
title('Calculated with polynomial assimilation code')
ylabel('BFL'); xlabel('Polynomial order')
subplot(3,1,2); plot(1:9,log_ML,'o-'); grid on
ylabel('ML'); xlabel('Polynomial order')
subplot(3,1,3); plot(1:9,log_OF,'o-'); grid on 
ylabel('OF'); xlabel('Polynomial order')

end



