function [a_MAP,inv_C_aa_MAP,log_ML,log_BFL,log_OF] = fun_MAP(N_mod,a_prior,inv_C_aa_prior,inv_C_ee,x_e,y_e)
% A demonstration of Bayesian parameter inference and model comparison
% based on polynomials.
%
% INPUTS:
%  N_mod            scalar      Order of the candidate polynomial model
%  a_prior          vector      prior polynomial coefficients
%  inv_C_aa_prior   matrix      inverse of the prior covariance matrix
%  inv_C_ee         scalar      inverse of the measurement variance
%  (x_e,y_e)        vectors     Data
%  plot_options     structure   plot options
%
% OUTPUTS:
%  a_MAP            vector      posterior polynomial coefficients
%  inv_C_aa_MAP     matrix      inverse of the posterior covariance matrix
%  log_ML           scalar      Log marginal likelihood of the candidate model
%  log_BFL          scalar      Log best fit likelihood of the candidate model
%  log_OF           scalar      Log Occam factor of the candidate model

%--------------------------------------------------------------------------
% Calculate the most likely parameter values (polynomial coefficients)
%--------------------------------------------------------------------------
%
% We estimate the most likely parameter values using the maximum a
% posteriori (MAP) method (MacKay pp 306). We find the set of
% parameters that maximise the posterior in Bayes theorem:
%
%            'Likelihood' 'Prior'
% 'Posterior'      |        |
%     |        P(Z|A,H) * P(A|H)
%  P(A|Z,H)  = -----------------                [1]
%                   P(Z|H)
%                     |
%       'Evidence' / 'Marginal Likelihood'
%
% It is more practical to minimise the negative log of the posterior, so we
% define this as our cost function (J). We also calculate the first and
% second derivatives of J w.r.t. the parameters. We then use a gradient
% based optimiser (Newton's method in this example) to minimise the cost
% function. 
%
% In addition to the posterior parameter values, we estimate the
% posterior parameter uncertainty using Laplace's method. We approximate
% the posterior distribution as a Gaussian, following which we see that the
% posterior covariance can be approximated as the Hessian of our cost
% function (MacKay ch. 27 and pp. 347).

% Initialise
tol      = 1e-10;
del_a    = 2*tol;
a        = a_prior; % starting point for the iteration

% Minimize J (the negative log posterior likelihood) using Newton's method
while norm(del_a) > tol    
    % Calculate dJ and d2J
    [J,dJ,d2J] = J_dJ_d2J(a,a_prior,inv_C_aa_prior,x_e,y_e,inv_C_ee);
    % Calculate del_a using Newton's method
    del_a = - d2J \ dJ.';
    % Update a
    a = a + del_a.';
end
% Re-label a and extract the inverse posterior covariance matrix
a_MAP        = a; inv_C_aa_MAP = d2J;

%--------------------------------------------------------------------------
% Calculate model ranking metrics
%--------------------------------------------------------------------------
% REMINDER:
%
%            'Likelihood' 'Prior'
% 'Posterior'      |        |
%     |        P(Z|A,H) * P(A|H)
%  P(A|Z,H)  = -----------------                [1]
%                   P(Z|H)
%                     |
%       'Evidence' / 'Marginal Likelihood'
%
% The evidence, P(Z,H) , equals int { P(Z|A,H) * P(A|H) } * dA  [Eq. 28.10 of MacKay], . 
% which can be estimated from Laplace's approximation:
%
% P(Z|H) ~ P(Z|A_MAP,H) * P(A_MAP,H) * det(inv_C_aa_MAP/2/pi)^(-0.5)
%   |           |         <---------------------------------------->
%   ML         BFL                              OF
%
% where ML  = marginal likelihood
%       BFL = best fit likelihood
%       OF  = Occam Factor
%
% Calculate log(P(A|Z,H)) (A = parameters, Z = data, H = hypothesis)
log_PAZH_MAP  = -J;
% Integrate P(Z|A,H)P(A|H) over parameter space A using Laplace's method (Eq 28.10 in MacKay) .
log_ML        = log_PAZH_MAP - 0.5 * log(det(inv_C_aa_MAP)) + 0.5 * (N_mod+1) * log(2*pi);
% Calculate log(P(A|H)) by re-calculating J without data
[J_prior,~,~] = J_dJ_d2J(a_MAP,a_prior,inv_C_aa_prior,[],[],inv_C_ee); log_PAH_MAP   = -J_prior;
% Calculate the BFL
log_BFL       = log_PAZH_MAP - log_PAH_MAP;
% Calculate the OF
log_OF        = log_ML - log_BFL;

end

%--------------------------------------------------------------------------
% Calculate the negative log likelihood and its sensitivities w.r.t. parameters  
%--------------------------------------------------------------------------
function [J,dJ,d2J] = J_dJ_d2J(a,a_prior,inv_C_aa,x_e,y_e,inv_C_ee)
% Calculate the cost function, (J) which is the negative log posterior likelihood of the parameters given the data: 
%
%           'Posterior'         'Likelihood'         'Prior'
%                |                   |                  |
%  J = - log( P(A|Z,H) ) = - log( P(Z|A,H) ) - log ( P(A|H) )
%                                   |                  |
%                                data Z(A)         parameters A   
%
%      <------ contribution from exponential ------->   <-- contribution from normalization -->
%  J = 0.5 * (a-a_prior)'   * inv_C_aa  * (a-a_prior)   + log( sqrt{ (2*pi)^N_a * det(C_aa)} ) [J1] <- contribution from prior discrepancy 
%    + 0.5 * (y_p(a)-y_e)'  * inv_C_eer * (y_p(a)-y_e)  + log( sqrt{ (2*pi)^n   * det(C_ee)} ) [J2] <- contribution from prediciton error 
%
% The first and second derivatives of the cost function with respect to the
% parameters are also calculated with a second order Taylor expansion. 
%
% For convenience, we drop the contributions from the normalization
% because they do not contain 'a', and we write J(a) in index form: 
%
%  J(a)  = 0.5 * (a-a_prior)_k  * inv_C_aa_kl * (a-a_prior)_l  [J1] 
%        + 0.5 * (y_p(a)-y_e)_i * inv_C_ee_ij * (y_p(a)-y_e)_j [J2] 
%
% Now we perform the second order Taylor expansion:
%
%  J(a0 + da) = 0.5 * (a0-a_prior)_k  * inv_C_aa_kl  * (a0-a_prior)_l  [J1] 
%             + 0.5 * (y_p(a0)-y_e)_i * inv_C_ee_ij  * (y_p(a0)-y_e)_j [J2] 
%           + {
%               (a0-a_prior)_k  * inv_C_aa_kl                          [dJ1/da_l]  }         
%             + (y_p(a0)-y_e)_i * inv_C_ee_ij * d(y_p)_j/d(a_l)|a0     [dJ2/da_l]  }  row vector dJ/dA       
%             } * da_l    
%           + 0.5 * da_k * {
%               inv_C_aa_kl                                                                                                             [d2J1/da_k/da_l]  }    
%             + d(y_p)_i/d(a_k)|a0 * inv_C_ee_ij * d(y_p)_j/d(a_l)|a0 + (y_p(a0)-y_e)_i  * inv_C_ee_ij * d2(y_p)_j/d(a_k)/d(a_l)|a0     [d2J2/da_k/da_l]  }  matrix d2J/dA/dA       
%             } * da_l
%
% This function outputs J, dJ/dA, and d2J/dA/dA

% Initialise
N_a = length(a);
J = 0; dJ = zeros(1,N_a); d2J = zeros(N_a);

% Calculate contribution of parameter prior discrepancy
J   = J   + 0.5 * (a-a_prior)    * inv_C_aa * (a-a_prior)';
J   = J   + 0.5 * N_a * log(2*pi) - sum(log(diag(chol(inv_C_aa))));
dJ  = dJ  + (a-a_prior)    * inv_C_aa;
d2J = d2J + inv_C_aa;

for ii = 1:length(x_e)
    % Evaluate current model predictions
    [y_p,dyda,d2ydada] = fun_solve_poly(a,x_e(ii));

    % Calculate cost contribution of prediction error
    J   = J   + 0.5 * (y_p-y_e(ii))'  * inv_C_ee * (y_p-y_e(ii));
    J   = J   + 0.5 * ( log(2*pi) - log(inv_C_ee) );
    dJ  = dJ  + (y_p-y_e(ii))  * inv_C_ee * dyda.' ;
    d2J = d2J + dyda * dyda.'  * inv_C_ee +...
        + (y_p-y_e(ii))  * inv_C_ee * d2ydada; % Always zero, but included for completeness
end
  
end
